SystemOrganization addCategory: #Squeak2JS!

TestCase subclass: #S2JParserTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Squeak2JS'!

!S2JParserTest methodsFor: 'helpers' stamp: 'ag 11/21/2010 11:58'!
assert: aString matches: aRule
	"In most cases, throws an error when failing (hard fail)"
	
	| result |
	result := S2JParser matchString: aString with: aRule.
	self assert: result notNil.
	^ result! !

!S2JParserTest methodsFor: 'helpers' stamp: 'ag 11/21/2010 12:00'!
assert: aString matches: aRule andProduces: anOrderedCollection
	"In most cases, throws an error when failing (hard fail)"
	
	| result |
	result := self assert: aString matches: aRule.
	self assert: result = anOrderedCollection.
	^ result! !

!S2JParserTest methodsFor: 'test cases' stamp: 'ag 11/21/2010 12:25'!
testBinarySelector
	
	self
		"special cases: single pipe and single minus"
		assert: '|' matches: #binarySelector andProduces: { S2JNode named: 'binarySelector' with: { '|' }};
		assert: '-' matches: #binarySelector andProduces: {};
		
		"names starting with a minus"
		assert: '-*' matches: #binarySelector andProduces: {};
		assert: '-***' matches: #binarySelector andProduces: {};
		assert: '-*%&' matches: #binarySelector andProduces: {};
		assert: '-*/\' matches: #binarySelector andProduces: {};
		assert: '-!!?' matches: #binarySelector andProduces: {};
		assert: '-==' matches: #binarySelector andProduces: {};
		
		"mixed characters, check each one"
		assert: '*+' matches: #binarySelector andProduces: {};
		assert: '/=\' matches: #binarySelector andProduces: {};
		assert: '~,,,<' matches: #binarySelector andProduces: {};
		assert: '>@' matches: #binarySelector andProduces: {};
		assert: '%&' matches: #binarySelector andProduces: {};
		assert: '?!!' matches: #binarySelector andProduces: {};
		assert: '``' matches: #binarySelector andProduces: {};
		assert: '+??=' matches: #binarySelector andProduces: {};
		assert: ',!!!!,' matches: #binarySelector andProduces: {}.! !

OMeta2 subclass: #S2JParser
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Squeak2JS'!

!S2JParser class methodsFor: 'matching' stamp: 'ag 11/18/2010 12:16'!
matchString: aString with: aRule
	
	| matcher result |
	matcher := self matcherOn: aString readStream.
	result := matcher
		apply: #empty;
		apply: aRule.
	matcher apply: #end.
	^ result! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/21/2010 13:47'!
binaryExpression =
	
	(primary | unaryExpression | binaryExpression):receiver spaces binarySelector:selector spaces (primary | unaryExpression):argument
	-> [ self node: 'binaryExpression' with: { receiver. selector. argument} ]! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/21/2010 13:46'!
binarySelector =
	
	<$| | ((specialCharacter | $-) (specialCharacter* | empty))>:selector
	-> [ selector asSymbol ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/21/2010 13:58'!
block =
	
	$[
		// arguments
		(((char($:) identifier:argumentName -> [ argumentName ])+:arguments char($|) ) | empty)
		// block contents
		spaces
		(temporaries:temporaries | empty)
		spaces
		(statements:statements | empty)
	char($])
	
	-> [ self node: 'block' withoutNils: {
			S2JNode named: 'arguments' withoutNilsNotEmpty: arguments.
			temporaries.
			statements} ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/21/2010 14:03'!
braceExpression =
	
	${ spaces
	((expression:firstExpression
		((char($.) spaces expression:expr -> [ expr ])*:expressions | empty))
	| empty)
	char($})
	-> [ self braceExpressionWith: firstExpression and: expressions ]! !

!S2JParser methodsFor: 'helpers' stamp: 'ag 11/21/2010 13:57'!
braceExpressionWith: firstExpression and: expressions
	
	^ self node: 'braceExpression' withoutNils: 
		((expressions
			ifNotNil: [ expressions ]
			ifNil: [ {} ])
				addFirst: firstExpression;
				yourself)! !

!S2JParser methodsFor: 'helpers' stamp: 'ag 11/21/2010 13:44'!
char :c =
	// Reads prepended spaces and returns only the character
	
	spaces <exactly(c)>:result
	-> [ result ]! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/21/2010 13:46'!
character =
	
	$$ <anything>:value
	-> [ value asCharacter ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/21/2010 14:04'!
expression =
	
	// TODO not fixed
	
	// Variable assignments
	(((identifier:identifier (token(':=') | token('_'))) -> [ identifier ])*:variableNames | empty)
	// Actual part of expression
	(messageExpression | primary ):expressionBody
	
	// TODO cascaded-message-expression!!
	// | cascadedMessageExpression
	
	// Create an extra node for the assignments, if any
	-> [ self node: 'expression' withoutNils: { 
			(self node: 'variableAssignments' withoutNilsNotEmpty: variableNames).
			expressionBody} ]! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/21/2010 13:45'!
identifier =
	// Does not read prepended spaces!!
	
	<letter (letter | digit)*>! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/21/2010 13:15'!
keywordSelector =
	
	<identifier $:>:selector
	-> [ selector asSymbol ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/18/2010 15:16'!
messageExpression =
	
	unaryExpression
|	binaryExpression
	
	// TODO keyword-Expressions!!
	//	|	keywordExpression
! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/18/2010 13:10'!
messagePattern =
	
	// Unary
	identifier:unarySelector -> [ S2JNode named: 'unary' with: { unarySelector } ]
	
	// Binary
|	binarySelector:selector identifier:argumentName -> [ S2JNode named: 'binary' with: { selector. argumentName } ]
	
	// Keyword
	// TODO
! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/18/2010 13:12'!
method =
	
	messagePattern:a1 (temporaries:a2 | empty) (primitiveDeclaration:a3 | empty) (statements:a4 | empty)
	-> [ S2JNode named: 'method' withoutNils: { a1. a2. a3. a4 } ]! !

!S2JParser methodsFor: 'helpers' stamp: 'ag 11/18/2010 23:33'!
node: aString with: anOrderedCollection
	
	^ S2JNode named: aString with: anOrderedCollection! !

!S2JParser methodsFor: 'helpers' stamp: 'ag 11/18/2010 23:33'!
node: aString withoutNils: anOrderedCollection
	
	^ S2JNode named: aString withoutNils: anOrderedCollection! !

!S2JParser methodsFor: 'helpers' stamp: 'ag 11/18/2010 23:34'!
node: aString withoutNilsNotEmpty: anOrderedCollection
	
	^ S2JNode named: aString withoutNilsNotEmpty: anOrderedCollection! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/21/2010 13:46'!
number =
	
	<(anything:n ?[n isDigit])+>:result
	-> [ result asNumber ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/21/2010 12:45'!
primary =
	// something, that evaluates to an object
	
	// TODO constant-array-literal; will need to outsource symbol-, number- and character-literals
	
	identifier // variableName / argumentName
|	number:number -> [ S2JNode named: 'number' with: { number } ]
|	symbolToken:symbolToken -> [ S2JNode named: 'symbol' with: { symbolToken } ]
|	character:character -> [ S2JNode named: 'character' with: { character }]
|	block
|	braceExpression
|	token('(') expression:expr token(')') -> [ expr ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/18/2010 13:15'!
primitiveDeclaration =
	
	// TODO primitive can also take the form <primitive: methodNameString module: moduleNameString>
	
	token('<') token('primitive:') digits:primitiveNumber token('>')
	-> [ S2JNode named: 'primitive' with: { primitiveNumber } ]! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/21/2010 13:46'!
specialCharacter =
	
	$+ | $* | $/ | $\ | $~ | $< | $> | $= | $@ | $% | $& | $? | $!! | $` | $,! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/19/2010 00:54'!
statements =
	
	// Some expressions
	(expression:expr token('.') -> [ expr ] )*:expressions
	
	// The last expression (return statement, not necessary, must not be followed by another expression)
	(((token('^'):returnSign | empty) expression:lastExpression (token('.') | empty) ~expression) | empty)
	
	// Helper method handles the resulting objects
	-> [ self statementsWithExpressions: expressions lastExpression: lastExpression returnSign: returnSign ]! !

!S2JParser methodsFor: 'helpers' stamp: 'ag 11/19/2010 00:57'!
statementsWithExpressions: expressions lastExpression: lastExpression returnSign: returnSign
	"Build the contents of a statements-node from the information gathered while parsing 'statements'"
	
	"Rename the last expression, if it was a return-expression"
	lastExpression notNil & returnSign notNil ifTrue: [ lastExpression name: 'returnExpression' ].
	expressions add: lastExpression.
	^ self node: 'statements' withoutNils: expressions! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/21/2010 13:46'!
string =
	
	$' <(~$' anything)*>:value $'
	-> [ value ]! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/21/2010 13:47'!
symbol =
	
	$# (identifier | string):value
	-> [ value asSymbol ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/18/2010 13:13'!
temporaries =
	
	token('|') identifier*:variableNames token('|')
	-> [ S2JNode named: 'temporaries' with: variableNames ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/18/2010 17:56'!
unaryExpression =
	
	(primary | unaryExpression):receiver identifier:message
	-> [ S2JNode named: 'unaryExpression' with: {receiver. message} ]! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/21/2010 13:13'!
unarySelector =
	
	identifier:selector
	-> [ selector asSymbol ]! !

OMeta2 subclass: #S2JTranslator
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Squeak2JS'!

!S2JTranslator methodsFor: 'as yet unclassified' stamp: 'ag 11/21/2010 12:12'!
node :aString =
	
	// This does not create an S2JNode, but reads one from the input stream and returns it!!
	// For this purpose, this is an ometa-rule with one argument, call like this: node('binarySelector')
	
	anything:node ?[node class = S2JNode and: [node name = aString]]
	-> [ node ]! !

OrderedCollection subclass: #S2JNode
	instanceVariableNames: 'name'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Squeak2JS'!

!S2JNode commentStamp: 'ag 11/18/2010 13:09' prior: 0!
This is only a named collection, that is supposed to hold further collections or strings to form a tree!

!S2JNode class methodsFor: 'instance creation' stamp: 'ag 11/18/2010 13:08'!
named: aString with: anOrderedCollection
	
	^ self new
		name: aString;
		addAll: anOrderedCollection;
		yourself! !

!S2JNode class methodsFor: 'instance creation' stamp: 'ag 11/18/2010 13:58'!
named: aString withoutNils: anOrderedCollection
	
	anOrderedCollection ifNil: [ ^ nil ].
	^ self
		named: aString
		with: (anOrderedCollection reject: [ :each | each isNil])! !

!S2JNode class methodsFor: 'instance creation' stamp: 'ag 11/18/2010 13:57'!
named: aString withoutNilsNotEmpty: anOrderedCollection
	"Returns nil, if resulting node would be empty"
	
	| result |
	result := self named: aString withoutNils: anOrderedCollection.
	^ (result isNil | result isEmpty)
		ifTrue: [ nil ]
		ifFalse: [ result ]! !

!S2JNode methodsFor: 'accessors' stamp: 'ag 11/21/2010 12:09'!
= anObject
	
	| result |
	result := super = anObject.
	result := result & (name = anObject name).
	^ result! !

!S2JNode methodsFor: 'accessors' stamp: 'ag 11/18/2010 13:07'!
name
	
	^ name! !

!S2JNode methodsFor: 'accessors' stamp: 'ag 11/18/2010 13:07'!
name: aString
	
	name := aString! !

!S2JNode methodsFor: 'accessors' stamp: 'ag 11/19/2010 00:39'!
printString
	
	| result first |
	first := true.
	result := name, ' ('.
	self do: [ :each |
		first ifTrue: [first := false] ifFalse: [result := result, ', '].
		result := result,  each printString].
	^ result, ')'! !
